select identity_type('authentication_subject_id', type => 'uuid', nextval => 'uuidv7');

create table authentication_subjects
(
    id authentication_subject_id not null primary key default authentication_subject_id_nextval()
);

select pg_catalog.pg_extension_config_dump('authentication_subjects', '');

comment on table authentication_subjects is $$
Authentication subject is a provisional term for subject of authentication, such as user.

Note, however, that we treat it a bit more broadly. For example, an unrecognized identifier (such as login
or e-mail) can also be an authentication subject. This way we can track attempts to authenticate against
a non-existent user.

This can also be useful in the context where we are doing an authentication for a user that does not yet exist,
especially in the context of third-party OAuth authentications-as-signups.
$$;
create domain password varchar(255);

create type hashing_algorithm as enum ('bcrypt');

create domain hashed_password text check (
    -- bcrypt
    value ~ '^\$2[aby]\$[0-9]{2}\$[./A-Za-z0-9]{53}$'
    );

create function work_factor(hashed_password hashed_password) returns int
    language sql
    immutable parallel safe as
$$
select case
           when omni_auth.hashing_algorithm(hashed_password) = 'bcrypt' then
               split_part(hashed_password, '$', 3)::int
           else null
           end
$$;
create function hashing_algorithm(hashed_password hashed_password) returns hashing_algorithm
    language sql
    immutable parallel safe as
$$
select case
           when hashed_password ~ '^\$2[aby]\$' then
               'bcrypt'
           else null::omni_auth.hashing_algorithm
           end
$$;

create function hash_password(password password, hashed_password hashed_password default null,
                              hashing_algorithm hashing_algorithm default 'bcrypt',
                              work_factor integer default null)
    returns hashed_password
    language plpgsql
    parallel safe
as
$$
declare
    _hashed_password omni_auth.hashed_password;
begin
    if hashed_password is not null then
        hashing_algorithm := omni_auth.hashing_algorithm(hashed_password);
        work_factor := omni_auth.work_factor(hashed_password);
    end if;
    if hashing_algorithm = 'bcrypt' then
        work_factor :=
                coalesce(work_factor, coalesce(current_setting('omni_auth.bcrypt_work_factor', true), 12::text)::int);
        -- we call `crypt` and `gen_salt` through `public` because `hash_password` can be called from `refresh materialized view`
        -- for `password_work_factor_timings` and the search_path will be restricted as of Postgres 17
        _hashed_password := public.crypt(password, coalesce(hashed_password, public.gen_salt('bf'::text, work_factor)));
        if hashed_password is not null and _hashed_password != hashed_password then
            return null;
        end if;
        return _hashed_password;
    else
        raise exception 'Unknown or missing hashing_algorithm';
    end if;
end;
$$;

select identity_type('password_credential_id', type => 'uuid', nextval => 'uuidv7');

create table password_credentials
(
    id password_credential_id not null default password_credential_id_nextval(),
    authentication_subject_id authentication_subject_id not null references authentication_subjects (id),
    hashed_password hashed_password not null,
    valid_at        tstzrange not null default tstzrange(statement_timestamp(), 'infinity'),
    constraint password_credentials_valid_at_no_overlap exclude using gist ((authentication_subject_id::uuid) with =, valid_at with &&)
);

create function set_password(authentication_subject_id authentication_subject_id, password password,
                             old_password password default null,
                             valid_from timestamptz default statement_timestamp(),
                             valid_until timestamptz default null,
                             hashing_algorithm hashing_algorithm default 'bcrypt',
                             work_factor integer default null)
    returns password_credentials
    language plpgsql
as
$$
declare
    result           omni_auth.password_credentials;
    _hashed_password omni_auth.hashed_password;
begin
    _hashed_password :=
            omni_auth.hash_password(password, hashing_algorithm => hashing_algorithm, work_factor => work_factor);
    with existing_credential as (select *
                                 from omni_auth.password_credentials
                                 where password_credentials.authentication_subject_id =
                                       set_password.authentication_subject_id
                                   and valid_at @> statement_timestamp()),
         credential as (select omni_auth.password_credential_id_nextval(),
                               existing_credential.authentication_subject_id,
                               case
                                   when old_password is null then _hashed_password
                                   when old_password is not null and omni_auth.hash_password(password => old_password,
                                                                                             hashed_password => hashed_password) is not null
                                       then
                                       _hashed_password
                                   else
                                       null
                                   end,
                               tstzrange(valid_from, coalesce(valid_until, upper(valid_at)))
                        from existing_credential
                        union
                        select omni_auth.password_credential_id_nextval(),
                               authentication_subject_id,
                               _hashed_password,
                               tstzrange(valid_from, coalesce(valid_until, 'infinity'))
                        where not exists (select from existing_credential)),
         update_validity
             as (update omni_auth.password_credentials set valid_at = tstzrange(lower(password_credentials.valid_at),
                                                                                valid_from) from existing_credential where password_credentials.id = existing_credential.id returning password_credentials.*),
         update as (select *
                    from update_validity
                    union
                    select (null::omni_auth.password_credentials).*
                    where not exists (select from update_validity))
    insert
    into omni_auth.password_credentials
    select credential.*
    from credential,
         update
    returning password_credentials.* into result;
    return result;
exception
    when not_null_violation then
        raise exception 'incorrect old_password';
end;
$$;

comment on table password_credentials is $$
Password credentials are temporally-constrained hashed passwords for authentication subjects
$$;

select pg_catalog.pg_extension_config_dump('password_credentials', '');

select identity_type('password_authentication_attempt_id', type => 'uuid', nextval => 'uuidv7');

--- We store password authentication attempts for detecting suspicious behavior
--- Example: HIPAA § 164.308(a)(5)(ii)(C)

create table password_authentication_attempts
(
    id                        password_authentication_attempt_id not null primary key default password_authentication_attempt_id_nextval(),
    authentication_subject_id authentication_subject_id references authentication_subjects (id),
    hashed_password        hashed_password,
    success     boolean   not null,
    occurred_at timestamp not null default statement_timestamp()
);

select pg_catalog.pg_extension_config_dump('password_authentication_attempts', '');

create function authenticate(password password, auth_subject_id authentication_subject_id,
                             as_of timestamptz default statement_timestamp())
    returns password_authentication_attempts
    language plpgsql
as
$$
declare
    result omni_auth.password_authentication_attempts;
begin
    with credentials as (select hashed_password
                         from omni_auth.password_credentials
                         where valid_at @> as_of
                           and authentication_subject_id = auth_subject_id),
         validation as (select auth_subject_id as subject_id,
                               hashed_password,
                               case
                                   when omni_auth.hashing_algorithm(hashed_password) = 'bcrypt' then
                                       crypt(password, hashed_password)
                                   -- unknown algorithm, pretend we're still doing hashing
                                   else omni_auth.hash_password(password)
                                   end as hashed_attempted_password
                        from credentials
                        -- otherwise, pretend to do work
                        union
                        select null as subject_id, '', omni_auth.hash_password(password)
                        where not exists (select from credentials))
    insert
    into omni_auth.password_authentication_attempts (authentication_subject_id, hashed_password, success)
    select subject_id,
           hashed_attempted_password,
           hashed_attempted_password = hashed_password
    from validation
    returning * into result;
    return result;
end
$$;

create function successful_authentication(password_authentication_attempts)
    returns boolean
    language sql
    immutable parallel safe
as
$$
select ($1).success
$$;

-- Informational section

--- To determine appropriate work factor, it would be useful to have some information on how long does it take to hash
--- for any given algorithm and work factor. OWASP suggestion is to target under 1s for hashing to ensure decent user experience
--- (https://cheatsheetseries.owasp.org/cheatsheets/Password_Storage_Cheat_Sheet.html#using-work-factors)

create function password_work_factor_timings(iterations int default 3, timeout_ms double precision default 1500)
    returns table
            (
                algorithm       hashing_algorithm,
                work_factor     int,
                iteration       int,
                time_elapsed_ms double precision
            )
    language plpgsql
    parallel safe
as
$$
declare
    current_work_factor int;
    current_iteration   int;
    start_time          timestamp;
    end_time            timestamp;
begin
    -- bcrypt
    algorithm := 'bcrypt';

    for current_work_factor in 10..31
        loop
            for current_iteration in 1..iterations
                loop
                    work_factor := current_work_factor;
                    iteration := current_iteration;
                    start_time := clock_timestamp();
                    perform omni_auth.hash_password('password', hashing_algorithm => algorithm,
                                                    work_factor => current_work_factor);
                    end_time := clock_timestamp();
                    time_elapsed_ms := (extract(epoch from end_time) - extract(epoch from start_time)) * 1000;
                    if time_elapsed_ms >= timeout_ms then
                        return;
                    end if;
                    return next;
                end loop;
        end loop;
    return;

end;
$$;

create materialized view password_work_factor_timings as
select *
from password_work_factor_timings() with no data;

create unique index password_work_factor_timings_idx on password_work_factor_timings (algorithm, work_factor, iteration);


