create function instantiate(schema regnamespace default 'omni_credentials',
                            env_var text default 'OMNI_CREDENTIALS_MASTER_PASSWORD') returns void
    language plpgsql
as
$$
begin
    perform set_config('search_path', schema::text || ',public', true);

    create table credentials_config
    (
        environment_variable text not null,
        unique (environment_variable)
    );

    insert into credentials_config (environment_variable) values (env_var);

    create table encrypted_credentials
    (
        name  text  not null unique,
        value bytea not null
    );

    create function credentials_master_password() returns text
        language sql
    as
    $sql$
    select value
    from omni_os.env
             inner join credentials_config on true
    where variable = environment_variable
    $sql$;
    execute format('alter function credentials_master_password set search_path to %I', schema);


    create view credentials as
    select name,
           pgp_sym_decrypt(value, credentials_master_password()) as value
    from encrypted_credentials;

    create function credentials_update() returns trigger
        security definer
        language plpgsql as
    $code$
    begin
        if old is distinct from null then
            update encrypted_credentials
            set value = pgp_sym_encrypt(new.value, credentials_master_password()),
                name  = new.name
            where name = old.name;
        else
            insert
            into encrypted_credentials (name, value)
            values (new.name, pgp_sym_encrypt(new.value, credentials_master_password()));
        end if;
        return new;
    end;
    $code$;
    execute format('alter function credentials_update set search_path to %I,public', schema);

    create trigger credentials_update
        instead of update or insert
        on credentials
        for each row
    execute function credentials_update();

end;
$$;
create function instantiate_file_store(filename text, schema regnamespace default 'omni_credentials') returns void
    language plpgsql
as
$$
begin
    perform set_config('search_path', schema::text || ',public', true);

    if filename not like '/%' then
        filename := current_setting('data_directory') || '/' || filename;
    end if;

    create table if not exists credential_file_stores
    (
        filename text unique
    );

    create or replace function credential_file_store_reload(filename text) returns boolean
        language plpgsql
    as
    $code$
    begin
        if filename not like '/%' then
            filename := current_setting('data_directory') || '/' || filename;
        end if;
        perform pg_stat_file(filename);
        create temp table __new_encrypted_credentials__
        (
            like encrypted_credentials
        ) on commit drop;
        execute format('copy __new_encrypted_credentials__ from %L', filename);

        insert into encrypted_credentials (name, value)
        select name, value
        from __new_encrypted_credentials__
        on conflict (name) do update set value = excluded.value;
        return true;
    exception
        when others then return false;
    end;
    $code$;
    execute format('alter function credential_file_store_reload set search_path to %I,public', schema);

    insert into credential_file_stores (filename) values (instantiate_file_store.filename);

    perform credential_file_store_reload(filename);
    execute format('copy encrypted_credentials to %L', filename);

    create or replace function file_store_credentials_update() returns trigger
        security definer
        language plpgsql as
    $code$
    declare
        rec record;
    begin
        for rec in select * from credential_file_stores
            loop
                execute format('copy encrypted_credentials to %L', rec.filename);
            end loop;
        return new;
    end;
    $code$;
    execute format('alter function file_store_credentials_update set search_path to %I,public', schema);

    perform
    from pg_trigger
    where tgname = 'file_store_credentials_update'
      and tgrelid = 'encrypted_credentials'::regclass;

    if not found then
        create constraint trigger file_store_credentials_update
            -- TODO: truncate can't be supported at this level
            after update or insert or delete
            on encrypted_credentials
            deferrable initially deferred
            for each row
        execute function file_store_credentials_update();
    end if;

end;
$$;
select instantiate(schema => 'omni_credentials');

