-- This is an amount type, representing non-negative amounts,
-- temporarily implemented as a domain. TODO: make it an unsigned
-- (pending omni_int release)
create domain amount as bigint check (value >= 0);

select identity_type('ledger_id', type => 'uuid', nextval => 'uuidv7');

create table ledgers
(
    id ledger_id primary key default ledger_id_nextval()
);

select pg_catalog.pg_extension_config_dump('ledgers', '');

select identity_type('account_id', type => 'uuid', nextval => 'uuidv7');

create table accounts
(
    id                               account_id primary key default account_id_nextval(),
    ledger_id                        ledger_id not null references ledgers (id),
    debits_allowed_to_exceed_credits boolean   not null     default false,
    credits_allowed_to_exceed_debits boolean   not null     default false,
    closed                           boolean   not null     default false,
    check ( credits_allowed_to_exceed_debits or debits_allowed_to_exceed_credits)
);

select pg_catalog.pg_extension_config_dump('accounts', '');

create table account_balance_slots
(
    account_id account_id not null references accounts (id),
    debited    amount     not null default 0,
    credited   amount     not null default 0,
    slot       int        not null,
    constraint account_slot_unique unique (account_id, slot)
);

select pg_catalog.pg_extension_config_dump('account_balance_slots', '');

create view account_balances as
select account_id,
       sum(debited) as debited,
       sum(credited)           as credited,
       sum(credited - debited) as balance
from account_balance_slots
group by account_id;

select identity_type('transfer_id', type => 'uuid', nextval => 'uuidv7');

create table transfers
(
    id                transfer_id primary key default transfer_id_nextval(),
    debit_account_id  account_id not null references accounts (id),
    credit_account_id account_id not null references accounts (id),
    amount amount not null
);

select pg_catalog.pg_extension_config_dump('transfers', '');

create function transaction_affected_accounts()
    returns table
            (
                account_id account_id,
                ledger_id  ledger_id
            )
    language c
as
'MODULE_PATHNAME';

create function statement_affected_accounts()
    returns table
            (
                account_id account_id,
                ledger_id  ledger_id
            )
    language c
as
'MODULE_PATHNAME';


create function deny_operation_on_immutable_table() returns trigger as
$$
begin
    raise exception 'records in this table are immutable';
end;
$$ language plpgsql;

create function deny_operation_on_accounts() returns trigger
    language c as
'MODULE_PATHNAME';

create function calculate_account_balances() returns trigger
    language c as
'MODULE_PATHNAME' stable;

create trigger calculate_account_balances
    after insert
    on transfers
    for each row
execute function calculate_account_balances();

create function update_account_balances() returns trigger
    language c as
'MODULE_PATHNAME' stable;

create trigger update_account_balances
    after insert
    on transfers
    for each statement
execute function update_account_balances();

create trigger immutable_transfers
    before delete or update or truncate
    on transfers
    for each statement
execute function deny_operation_on_immutable_table();

create trigger immutable_accounts
    before delete or update
    on accounts
    for each row
execute function deny_operation_on_accounts();

create trigger untruncatable_accounts
    before truncate
    on transfers
    for each statement
execute function deny_operation_on_immutable_table();

create function ensure_affected_accounts_balance() returns trigger
    language plpgsql
as
$$
declare
    balance omni_ledger.amount;
begin
    select sum(ab.balance)
    into balance
    from omni_ledger.account_balances ab
             inner join omni_ledger.statement_affected_accounts() tat
                        on tat.account_id = ab.account_id;
    if balance != 0 then
        raise exception 'violation: accounts did not balance: %', balance;
    end if;
    return null;
end;
$$;

create trigger balancing_accounts
    after insert
    on transfers
    for each statement
execute function ensure_affected_accounts_balance();

comment on trigger balancing_accounts on transfers is
    $$ A debugging assertion consistency check that ensures that every transfer is balanced between debit and credit side.
    Disabled by default as users can't violate this invariant, but omni_ledger, in theory, can.$$;

alter table transfers
    disable trigger balancing_accounts;
create type category_type as enum ('asset', 'liability', 'equity');
select identity_type('account_category_id');

create table account_categories
(
    id                 account_category_id not null primary key default account_category_id_nextval(),
    name               text                not null,
    type               category_type       not null,
    debit_normal boolean not null default false
);

select pg_catalog.pg_extension_config_dump('account_categories', '');

create or replace function account_categories_immutability() returns trigger as
$$
begin
    if new.type != old.type then
        raise exception 'cannot change the account category type once it has been set';
    end if;
    if new.debit_normal != old.debit_normal then
        raise exception 'cannot change account category debit normalness once it has been set';
    end if;
    return new;
end;
$$ language plpgsql;

create trigger account_categories_immutability
    before update
    on account_categories
    for each row
execute function account_categories_immutability();

create table account_categorizations
(
    account_id  account_id          not null references accounts (id),
    category_id account_category_id not null references account_categories (id) on delete restrict,
    unique (account_id, category_id)
);

select pg_catalog.pg_extension_config_dump('account_categorizations', '');

create trigger immutable_account_categorizations
    before delete or update or truncate
    on account_categorizations
    for each statement
execute function deny_operation_on_immutable_table();

create view accounting_balances as
select account_categorizations.account_id,
       account_categories.type    as category_type,
       credited,
       debited,
       balance * (case
                      when account_categories.debit_normal then -1
                      else 1 end) as balance
from account_categorizations
         inner join account_balances on account_balances.account_id = account_categorizations.account_id
         inner join account_categories on account_categories.id = account_categorizations.category_id;

create function ensure_affected_accounts_accounting_balances() returns trigger
    language plpgsql
as
$$
declare
    rec          record;
    _assets      numeric;
    _liabilities numeric;
    _equity      numeric;
    _ledger_id   omni_ledger.ledger_id;
begin
    _ledger_id := null;
    for rec in
        select sum(balance) balance, ledger_id, category_type
        from omni_ledger.accounting_balances ab
                 inner join omni_ledger.statement_affected_accounts() sac
                            on sac.account_id = ab.account_id
        group by sac.ledger_id, category_type
        loop
            if _ledger_id is null or rec.ledger_id != _ledger_id then
                _assets := 0;
                _liabilities := 0;
                _equity := 0;
                if _ledger_id is not null then
                    if _assets != _liabilities + _equity then
                        raise exception 'violation: accounting did not balance for ledger %: A (%) != L (%) + E (%)', _ledger_id, _assets, _liabilities, _equity;
                    end if;
                end if;
                _ledger_id := rec.ledger_id;
            end if;
            case
                when rec.category_type = 'asset' then _assets := _assets + rec.balance;
                when rec.category_type = 'liability' then _liabilities := _liabilities + rec.balance;
                when rec.category_type = 'equity' then _equity := _equity + rec.balance;
                end case;

        end loop;
    -- for the last ledger:
    if _assets != _liabilities + _equity then
        raise exception 'violation: accounting did not balance for ledger %: A (%) != L (%) + E (%)', _ledger_id, _assets, _liabilities, _equity;
    end if;
    return null;
end;
$$;

create trigger balancing_accounting
    after insert
    on transfers
    for each statement
execute function ensure_affected_accounts_accounting_balances();

